// ************************************************************************	//
//                                                                          //
// Program : RLE                                                            //
// Tasks(s): Take a file and either encode to run-length, or decode from.   //
// Programmer: Robert L. Bryant	(i.m.roth@gmail.com)                        //
//                                                                          //
// For anyone that may not know what run-length encoding is, here is a      // 
// nicely put example from wikipedia.org:                                   //
//  Run-length encoding (RLE) is a very simple form of data compression in  //
//  which runs of data (that is, sequences in which the same data value     //
//  occurs in many consecutive data elements) are stored as a single data   //
//  value and count, rather than as the original run.                       //
//                                                                          //
//  EXAMPLE:                                                                //
//                                                                          //
//  WWWWWWWWWWWWBWWWWWWWWWWWWBBBWWWWWWWWWWWWWWWWWWWWWWWWBWWWWWWWWWWWWWW     //
//                                                                          //
//  If we apply the run-length encoding (RLE) data compression algorithm    //
//  to the above hypothetical scan line, we get the following:              //
//                                                                          //
//  12W1B12W3B24W1B14W                                                      //
//                                                                          //
//  Interpret this as twelve W's, one B, twelve W's, three B's, etc."       //
//                                                                          //
// ************************************************************************ //

#include <iostream>
#include <fstream>
#include <sys/stat.h>
#include <string>

using namespace std;

struct stat results;        // To be used for getting the file size of the file for action

int main(int argc, char* argv[])
{
	int fileSize;                                   // Obviously to hold the size of the file ;)
	unsigned short int firstByte, secondByte;       // Variables of bytes to be read to/from the files
	unsigned short int runCounter = 1;              // Counter
	char size8bit;                                  // Use to keep size of output bytes to 8 bit
	string convertArg;                              // For converting argv[2] to string

	if (argc != 4)
	{
		cout << "\nUsage: " << argv[0] << " <File for action>";                     // If the user doesn't enter
		cout << " [option (!REQUIRED!)]" << " <File output name>" << endl;          // all arguments, output the
		cout << "Options:" << endl;                                                 // correct usage and display
		cout << "  -e        Apply run-length encoding" << endl;                    // the options
		cout << "  -d        Decode a file from run-length encoding\n" << endl;
	}
	else
	{
		ifstream inFile (argv[1], ios::in | ios::binary);                   // Open files in binary mode
		ofstream outFile (argv[3], ios::out | ios::binary);

		convertArg = argv[2];       // Here we convert the option to string,
                                    // in this way we can read the argument
                                    // inside the 'if'

// ************************************************************ //
//                                                              //
// Here we begin the code for encoding to run-length encoding   //
//                                                              //
// ************************************************************ //

		if (convertArg == "-e")
		{
			if (!inFile.is_open())                                          // If inFile didn't open properly,
				cout << "File does not exist." << endl;                     // let the user know
			else
			{
				if (stat(argv[1], &results) == 0)                           // Get the size of the file
					fileSize = results.st_size;                             // through use of stat.h

				firstByte = inFile.get();                                   // Read the first byte of the file

				for (int i = 0; i < fileSize; i++)                  // Start the loop, and keep going until
				{                                                   // fileSize is met		
					secondByte = inFile.get();                      // Grab the next byte in the file...

					if (firstByte == secondByte)                    // ... then check to see if the first
					{                                               // and second bytes are the same
						if (runCounter == 255)                      // If our counter for a byte reaches
						{                                           // the maximum...
							outFile.write((char*)& runCounter, sizeof size8bit);    // ... write the number of times
							outFile.write((char*)& firstByte, sizeof size8bit);     // the next byte iterates, then 
							firstByte = secondByte;                                 // place that byte next in the file
							runCounter = 1;                                         // Store secondByte to firstByte for
						}                                                           // comparison on the next iteration,
						else                                                        // and reset counter
						{
							runCounter++;        // If the counter isn't maxed, increment it
						}
					}

					else
					{
						outFile.write((char*)& runCounter, sizeof size8bit);    // If the bytes don't match,
						outFile.write((char*)& firstByte, sizeof size8bit);     // do the same as above in the
						firstByte = secondByte;                                 // runCounter == 255 'if'
						runCounter = 1;                                         // This probably could've been
					}                                                           // made into a function, but
                                                                                // I figured I would leave it
				}                                                               // as is for now

				inFile.close();          // Close the file streams
				outFile.close();

			}

		}

// ************************************************************ //
//                                                              //
// Here we begin the code for decoding from run-length encoding //
//                                                              //
// ************************************************************ //

		else
		{
			if (!inFile.is_open())                              // If inFile didn't open properly,
				cout << "File does not exist." << endl;         // let the user know
			else
			{
				if (stat(argv[1], &results) == 0)               // Get the size of the file
					fileSize = results.st_size;                 // through use of stat.h

				for (int z = 0; z < fileSize; z++)              // Our loop to read the file
				{
					firstByte = inFile.get();                   // Grab the two bytes, the first being
					secondByte = inFile.get();                  // length of the run, second being the
                                                                // data to place in the outFile
					for (int x = 0; x < firstByte; x++)         // This is the loop to actually have
					{                                           // the data go out to the file
						if (inFile.eof())                       // and keep writing until the end of
						{                                       // inFile is reached
							inFile.close();
							outFile.close();
							return 1;
						}
						else
							outFile.write((char*)& secondByte, sizeof size8bit);
					}

				}

			inFile.close();         // Just in case, though I don't think it needs to be here because of the
			outFile.close();        // inFile.eof() 'if' above

			}

		}

	}

	return 0;
}
